/*
 *  $Id: //depot/prod/rlm/Rsp1.3_rlm/netapp/tools/miitool/miitool.c#1 $
 * 
 *  This tool prints out the following ethernet information
 *  about the given interface:
 * 
 *  link status
 *  auto negotiation status
 *  link speed
 *  link duplex
 * 
 *  This program is based on David Hinds' mii-tool (version
 *  1.9.1.1), which is a much more general purpose tool. The
 *  source code for mii-tool is available at:
 *  http://www.sfr-fresh.com/unix/privat/mii-tool-1.9.1.1.tar.gz/
 * 
 *  Copyright (c) 2007 - Krishna Kotnana, Network Appliance Inc.
 * 
 * 
 *  mii-tool: monitor and control the MII for a network interface
 * 
 *    Usage:
 * 
 *  mii-tool [-VvRrw] [-A media,... | -F media] [interface ...]
 * 
 *    This program is based on Donald Becker's "mii-diag" program, which
 *    is more capable and verbose than this tool, but also somewhat
 *    harder to use.
 * 
 *    Copyright (C) 2000 David A. Hinds -- dhinds@pcmcia.sourceforge.org
 * 
 *    mii-diag is written/copyright 1997-2000 by Donald Becker
 *        <becker@scyld.com>
 * 
 *    This program is free software; you can redistribute it
 *    and/or modify it under the terms of the GNU General Public
 *    License as published by the Free Software Foundation.
 * 
 *    Donald Becker may be reached as becker@scyld.com, or C/O
 *    Scyld Computing Corporation, 410 Severn Av., Suite 210,
 *    Annapolis, MD 21403
 * 
 *    References
 *  http://www.scyld.com/diag/mii-status.html
 *  http://www.scyld.com/expert/NWay.html
 *  http://www.national.com/pf/DP/DP83840.html
 *
 */

#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <sys/ioctl.h>
#include <sys/socket.h>
#include <net/if.h>
#include "mii.h"

/*
 * SIOCGMIIPHY may be undeclared in some cases. Where it is
 * undefined (older kernels), we define it to SIOCDEVPRIVATE
 */
#ifndef SIOCGMIIPHY 
#define SIOCGMIIPHY SIOCDEVPRIVATE
#endif


/*
 * These definition make some bit comparisions easy.
 */
#define LPA_10             (LPA_10HALF | LPA_10FULL)
#define LPA_DUPLEX_FULL    (LPA_10FULL | LPA_100FULL)
#define LPA_DUPLEX_HALF    (LPA_10HALF | LPA_100HALF)


/*
 * Some type definitions definitions
 */
typedef enum {
	MII_ERR_SUCCESS,      // Success
	MII_ERR_INVALID_PARM, // Invalid input parameter
	MII_ERR_FILE_WRITE,   // Error writing to stdout
	MII_ERR_SHELL_FAIL,   // Shell command failed
	MII_ERR_SOCKET_OPEN,  // Error opening socket for the particular interface
	MII_ERR_BMSR_READ,    // Error reading BMSR reg
	MII_ERR_BMCR_READ,    // Error reading BMCR reg
	MII_ERR_ACR_READ,     // Error reading ACR reg
	MII_ERR_LPAR_READ,    // Error reading LPAR reg
	MII_ERR_UNKNOWN,      // Error unknown!
} mii_err_t;

/*
 * Link status
 */
typedef enum link_status {
	LINK_STATUS_UNKNOWN,
	LINK_STATUS_UP,
	LINK_STATUS_DOWN,
} link_status_t;

#define LINK_STATUS_STRING_UNKNOWN            "unknown"
#define LINK_STATUS_STRING_UP                 "up"
#define LINK_STATUS_STRING_DOWN               "down"


/*
 * Auto negotiation status
 */
typedef enum aneg_status {
	ANEG_STATUS_UNKNOWN,
	ANEG_STATUS_ENABLED_COMPLETE,   // Auto negotiation enabled and complete
	ANEG_STATUS_ENABLED_INCOMPLETE, // Auto negotiation enabled but not complete
	ANEG_STATUS_ENABLED_FAILED,     // Auto negotiation enabled and failed
	ANEG_STATUS_DISABLED,           // Auto negotiatino disabled
} aneg_status_t;

#define ANEG_STATUS_STRING_UNKNOWN            "unknown"
#define ANEG_STATUS_STRING_ENABLED_COMPLETE   "complete"
#define ANEG_STATUS_STRING_ENABLED_INCOMPLETE "incomplete"
#define ANEG_STATUS_STRING_ENABLED_FAILED     "failed"
#define ANEG_STATUS_STRING_DISABLED           "disabled"


/*
 * Link speed 
 */
typedef enum link_speed {
	LINK_SPEED_UNKNOWN, // Cannot determine speed
	LINK_SPEED_10,     // 10 Mbps  (could be half duplex or full duplex
	LINK_SPEED_100,    // 100 Mbps (could be half duplex, full duplex, base4
} link_speed_t;
#define LINK_SPEED_STRING_UNKNOWN              "unknown"
#define LINK_SPEED_STRING_10                   "10 Mbps"
#define LINK_SPEED_STRING_100                  "100 Mbps"

/*
 * Duplex
 */
typedef enum link_duplex {
	LINK_DUPLEX_UNKNOWN,
	LINK_DUPLEX_HALF,   
	LINK_DUPLEX_FULL,
} link_duplex_t;
#define LINK_DUPLEX_STRING_UNKNOWN              "unknown"
#define LINK_DUPLEX_STRING_HALF                 "half"
#define LINK_DUPLEX_STRING_FULL                 "full"



/*
 * Data structure for mii info
 */
typedef struct mii_info {
	link_status_t link_status;
	aneg_status_t aneg_status;
	link_duplex_t link_duplex;
	link_speed_t link_speed;
} mii_info_t;


struct mii_ioctl_data {
	uint phy_id;
	uint reg_num;
	uint val_in;
	uint val_out;
};

struct lookup {
	int key;
	char *val;
};

static const char * lookup(int key, struct lookup table[], int num_entries)
{	
	int i;
	for (i = 0; i < num_entries; i++) {
		if (table[i].key == key) {
			return table[i].val;
		}
	}
	return "invalid";
}


/*
 * Link status string look up.
 */
static const char* link_status_str(link_status_t status)
{
	struct lookup link_status_str_lookup[] = {
		{ LINK_STATUS_UNKNOWN, LINK_STATUS_STRING_UNKNOWN },
		{ LINK_STATUS_UP,      LINK_STATUS_STRING_UP      },
		{ LINK_STATUS_DOWN,    LINK_STATUS_STRING_DOWN    },
	};
	return lookup(status, 
		      link_status_str_lookup,
		      sizeof(link_status_str_lookup)/sizeof(link_status_str_lookup[0]));
}



/*
 * Auto  negotiation status string look up
 */
static const char* aneg_status_str(aneg_status_t status)
{
	struct lookup aneg_status_str_lookup[] = {
		{ ANEG_STATUS_UNKNOWN,             ANEG_STATUS_STRING_UNKNOWN            },
		{ ANEG_STATUS_ENABLED_COMPLETE,    ANEG_STATUS_STRING_ENABLED_COMPLETE   },
		{ ANEG_STATUS_ENABLED_INCOMPLETE,  ANEG_STATUS_STRING_ENABLED_INCOMPLETE },
		{ ANEG_STATUS_ENABLED_FAILED,      ANEG_STATUS_STRING_ENABLED_FAILED     },
		{ ANEG_STATUS_DISABLED,            ANEG_STATUS_STRING_DISABLED           },	
	};
	return lookup(status, 
		      aneg_status_str_lookup,
		      sizeof(aneg_status_str_lookup)/sizeof(aneg_status_str_lookup[0]));
}


/*
 * Link speed string lookup
 */
static const char* link_speed_str(link_speed_t speed)
{
	struct lookup link_speed_str_lookup[] = {
		{ LINK_SPEED_UNKNOWN, LINK_SPEED_STRING_UNKNOWN },
		{ LINK_SPEED_10,      LINK_SPEED_STRING_10      },
		{ LINK_SPEED_100,     LINK_SPEED_STRING_100     },
	};
	return lookup(speed,
		      link_speed_str_lookup,
		      sizeof(link_speed_str_lookup) / sizeof(link_speed_str_lookup[0]));
}

/*
 * Link duplex string lookup
 */
static const char* link_duplex_str(link_duplex_t duplex)
{
	struct lookup link_duplex_str_lookup[] = {
		{ LINK_DUPLEX_UNKNOWN, LINK_DUPLEX_STRING_UNKNOWN }, 
		{ LINK_DUPLEX_HALF,    LINK_DUPLEX_STRING_HALF    },
		{ LINK_DUPLEX_FULL,    LINK_DUPLEX_STRING_FULL    },
	};
	return lookup(duplex,
		      link_duplex_str_lookup,
		      sizeof(link_duplex_str_lookup) / sizeof(link_duplex_str_lookup[0]));
}


/*
 * mii_info_t to string conversion. Useful for printing
 * information. 
 */
static int mii_print_info(const mii_info_t *info, char *buf, int n)
{
	int rc = 0;
	rc += snprintf(buf,
		 n,
		 "link status: %s, auto negotiate: %s, link speed: %s, link duplex: %s",
		 link_status_str(info->link_status),
		 aneg_status_str(info->aneg_status),
		 link_speed_str(info->link_speed),
		 link_duplex_str(info->link_duplex));
	return rc;
}

/*
 * Fetch the following information about the interface specified
 * 1. Link status
 * 2. Auto negotiation status
 * 3. Link speed
 * 4. Link duplex
 * 
 */
static mii_err_t mii_get_info(char *ifname, mii_info_t *mii_info)
{
	int fd = -1;
	struct ifreq ifr;
	struct mii_ioctl_data * mii_data = (struct mii_ioctl_data *) &ifr.ifr_data;

	int bmcr; // Basic mode control register
	int bmsr; // Basic mode status register
	int acr;  // Advertisement control register
	int lpar; // Link partner ability register
	int neg;  // Negotiated result
	int rc = MII_ERR_SUCCESS;	  


	memset(&ifr, 0, sizeof(ifr));
	mii_info->link_status = LINK_STATUS_UNKNOWN;
	mii_info->link_duplex = LINK_DUPLEX_UNKNOWN;
	mii_info->link_speed = LINK_SPEED_UNKNOWN;
	mii_info->aneg_status = ANEG_STATUS_UNKNOWN;



	fd = socket(AF_INET, SOCK_STREAM, 0);
	if (fd < 0) {
		rc = MII_ERR_SOCKET_OPEN;
		goto finish;
	}

	snprintf(ifr.ifr_name, IFNAMSIZ, "%s", ifname);

	mii_data->reg_num = MII_BMSR;
	if (ioctl(fd, SIOCGMIIPHY, &ifr)) {
		rc = MII_ERR_BMSR_READ;
		goto finish;
	}
	bmsr = mii_data->val_out;

	mii_data->reg_num = MII_BMCR;
	if (ioctl(fd, SIOCGMIIPHY, &ifr)) {
		rc = MII_ERR_BMCR_READ;
		goto finish;
	}
	bmcr = mii_data->val_out;

	mii_data->reg_num = MII_ADVERTISE;
	if (ioctl(fd, SIOCGMIIPHY, &ifr)) {
		rc = MII_ERR_ACR_READ;
		goto finish;
	}
	acr = mii_data->val_out;

	mii_data->reg_num = MII_LPA;
	if (ioctl(fd, SIOCGMIIPHY, &ifr)) {
		rc = MII_ERR_LPAR_READ;
		goto finish;
	}
	lpar = mii_data->val_out;

	/*
         * Link status
         */
	mii_info->link_status = bmsr & BMSR_LSTATUS 
				? LINK_STATUS_UP 
				: LINK_STATUS_DOWN;


	/*
         * ANEG, DUPLEX and SPEED
         */
	if (bmcr & BMCR_ANENABLE) {
		if (bmsr & BMSR_ANEGCOMPLETE) {
			mii_info->aneg_status = ANEG_STATUS_ENABLED_COMPLETE;
			/*
			 * If ANEG is complete, then ACR and LPAR should tell us about
			 * speed and duplex
                         */
			neg = acr & lpar;
			mii_info->aneg_status = neg ? ANEG_STATUS_ENABLED_COMPLETE 
			                            : ANEG_STATUS_ENABLED_FAILED;

			/*
			 * It seems important that the higher speed case should be
			 * checked first. So if LPA_1000 is to be added it should be
			 * checked before LPA_100.
                         */
			if (neg & LPA_100) {
				mii_info->link_speed = LINK_SPEED_100;
			} else if (neg & LPA_10) {
				mii_info->link_speed = LINK_SPEED_10;
			}
				
			if (neg & LPA_DUPLEX_FULL) {
				mii_info->link_duplex = LINK_DUPLEX_FULL;
			} else if (neg & LPA_DUPLEX_HALF) {
				mii_info->link_duplex = LINK_DUPLEX_HALF;
			}

		} else {
			mii_info->aneg_status = ANEG_STATUS_ENABLED_INCOMPLETE;
		}
	} else {
		/*
                 * If ANEG is not enabled get speed and duplex from BMCR
                 */
		mii_info->aneg_status = ANEG_STATUS_DISABLED;
		if (bmcr & BMCR_SPEED100) {
			mii_info->link_speed = LINK_SPEED_100;
		} else {
			mii_info->link_speed = LINK_SPEED_10;
		}

		if (bmcr & BMCR_FULLDPLX) {
			mii_info->link_duplex = LINK_DUPLEX_FULL;
		} else {
			mii_info->link_duplex = LINK_DUPLEX_HALF;
		}
	}
finish:
	if (fd >= 0) {
		close(fd);
	}
	return rc;
}

int main(int argc, char *argv[])
{
	mii_info_t info;
	char buf[256];
	int rc;
	int binary_dump = 0;

	switch (argc) {
	case 2:
		break;
	case 3:
		if (!strcmp(argv[2], "bin")) {
			binary_dump = 1;
		} else {
			fprintf(stderr, "argument 2 invalid");  
			return MII_ERR_INVALID_PARM;
		}
		break;
	default:
		fprintf(stderr, "Usage: %s <ifname> \n for example: %s eth0\n", argv[0], argv[0]);
		return MII_ERR_INVALID_PARM;
	}

	if (strlen(argv[1]) >= IFNAMSIZ) {
		fprintf(stderr, "Invalid parameter: interface name too long\n");
		return MII_ERR_INVALID_PARM;
	}


	rc = mii_get_info(argv[1], &info);
	if (rc == MII_ERR_SUCCESS) {
		if (binary_dump) {
			if (fwrite(&info, sizeof(info), 1, stdout) != 1) {
				fprintf(stderr, "error writing to stdout");
				rc = MII_ERR_FILE_WRITE;
			}
		} else {
			mii_print_info(&info, buf, sizeof(buf));			
			printf("%s\n", buf);
		}
	} else {
		if (!binary_dump) {
			fprintf(stderr, "error fetching ethernet information. code = %d\n", rc);
		}
	}

	return rc;
}
